package forum;

/**
 * 
 * @author Peter Knees <peter.knees@jku.at>
 * 
 * License: CreativeCommons Attribution-NonCommercial-ShareAlike 4.0 International
 * http://creativecommons.org/licenses/by-nc-sa/4.0/
 *
 */

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.net.MalformedURLException;
import java.net.URL;
import java.net.URLConnection;
import java.nio.charset.Charset;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.StringTokenizer;
import java.util.Vector;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import java.util.HashSet;

import com.fasterxml.jackson.core.JsonParseException;
import com.fasterxml.jackson.core.type.TypeReference;
import com.fasterxml.jackson.databind.JsonMappingException;
import com.fasterxml.jackson.databind.ObjectMapper;

import beatport.jsonpojo.Entry;
import beatport.jsonpojo.StandardKey;
import cp.text.HTMLCharacters;
import cp.text.PlainTextExtractor;
import cp.text.TextTool;


public class ExtractForumGroundTruth {

	private static Pattern bpm_pattern = Pattern.compile("(^|[=\\p{Space}])[0-9]{2,3}([,\\.][0-9]{1,3})?");
	private static Pattern[] wrong_bpm_patterns = new Pattern[]{
		Pattern.compile("not\\p{Space}*[0-9]{2,3}"),
		Pattern.compile("wrong[\\:bpm\\p{Space}]{1,5}\\p{Space}*[0-9]{2,3}")
	};

	private static Pattern key_pattern = Pattern.compile("(^|[=\\p{Space}])[a-h](\\p{Space}*(#|b|sharp|flat))?(\\p{Space}*|$)((min|maj)(or)?)");
	private static Pattern wrong_key_pattern = Pattern.compile("(^|[=\\p{Space}])not\\p{Space}+[a-h](\\p{Space}*(#|b|sharp|flat))?(\\p{Space}*|$)((min|maj)(or)?)");

	private static ObjectMapper mapper = new ObjectMapper();
	
	private static boolean outputverbose = false;

	private static File forumdir = new File("./forumpages/");
	private static File jsondir = new File("./beatport_json/");

	public static void downloadForumPages() throws JsonParseException, JsonMappingException, IOException {
		int pageidx = 1;
		int totalnumcomments = 0;
		int numcomments = 0;
		forumdir.mkdirs();
		jsondir.mkdirs();
		while (true) {

			String pagecontent = null;
			try {
				pagecontent = retrieve(
						new URL("https://beatport.uservoice.com/forums/125821-beatport/suggestions/2097707-fix-bpm-key-for-the-following-tracks?page="+(pageidx)+"&per_page=20")
						);
				TextTool.stringToFile(pagecontent, new File(forumdir.getAbsolutePath()+"/forumpage_"+pageidx+".html"), "utf-8");
			} catch (MalformedURLException e) {
				e.printStackTrace();
				return;
			}
			pagecontent = pagecontent.substring(0, pagecontent.indexOf("<div class=\"uvPagination\">"));
			String[] comments = pagecontent.split("<article class=\"uvUserAction uvUserAction-comment\">");
			if (comments.length <= 1)
				break;

			for (int i = 1; i < comments.length; i++) {

				//System.out.println(urls[i]);
				//int end = urls[i].indexOf("\\\" onmousedown");
				int end = comments[i].indexOf("</article>");
				if (end==-1)
					continue;
				totalnumcomments++;
				String commentblock = comments[i].substring(0, end).trim();
				int commentstart = commentblock.indexOf("<div class=\"typeset\">") + 21;
				int commentend = commentblock.indexOf("</div>", commentstart);
				String comment = commentblock.substring(commentstart, commentend);
				URL[] urls = extractTrackLinks(comment);
				comment = PlainTextExtractor.extractPlainText(removeLinks(comment));

				if (outputverbose) System.out.println("COMMENT (on "+pageidx+") "+comment);

				if (urls.length > 0) {
					numcomments++;
					if (outputverbose) System.out.println(" URL "+urls[0].toString());
					// extract trackid
					int idstart = urls[0].toString().lastIndexOf("/")+1;
					String trackid = urls[0].toString().substring(idstart, urls[0].toString().length());
					// must be number
					try {
						Integer.parseInt(trackid);
					} catch (NumberFormatException e) {
						//e.printStackTrace();
						continue;
					}
					if (outputverbose) System.out.println(" TRACKID "+trackid);

					String linkcontent = retrieve(urls[0]);			
					String[] jsons = linkcontent.split("data-json='");
					for (int j = 0; j < jsons.length; j++) {
						int jsonend = jsons[j].indexOf("'><a class");
						if (jsonend==-1)
							continue;
						String jsoninfo = jsons[j].substring(0, jsonend).trim();
						Entry entry = mapper.readValue(jsoninfo, new TypeReference<Entry>() {});
						String jsonid = entry.getId();
						if (jsonid.equals(trackid))
							TextTool.stringToFile(jsoninfo, new File(jsondir.getAbsolutePath()+"/"+jsonid+".json"), "utf-8");
					}
				}
			}
			pageidx++;
		}
		if (outputverbose) System.out.println("no. comments total: "+totalnumcomments+" / with URL:"+numcomments+"\n");

	}

	public static void extractGroundTruth() throws JsonParseException, JsonMappingException, IOException {

		int pageidx = 1;
		int totalnumcomments = 0;
		int numcomments = 0;
		int numbpm = 0;
		int numkeys = 0;
		forumdir.mkdirs();
		jsondir.mkdirs();

		while (true) {

			String pagecontent;
			try {
				pagecontent = TextTool.readInFile(new File(forumdir.getAbsolutePath()+"/forumpage_"+pageidx+".html"), "utf-8");
			} catch (Exception e) {
				e.printStackTrace();
				return;
			}
			pagecontent = pagecontent.substring(0, pagecontent.indexOf("<div class=\"uvPagination\">"));
			String[] comments = pagecontent.split("<article class=\"uvUserAction uvUserAction-comment\">");
			if (comments.length <= 1)
				break;
			for (int i = 1; i < comments.length; i++) {

				//System.out.println(urls[i]);
				//int end = urls[i].indexOf("\\\" onmousedown");
				int end = comments[i].indexOf("</article>");
				if (end==-1)
					continue;
				totalnumcomments++;
				String commentblock = comments[i].substring(0, end).trim();
				int commentstart = commentblock.indexOf("<div class=\"typeset\">") + 21;
				int commentend = commentblock.indexOf("</div>", commentstart);
				String comment = commentblock.substring(commentstart, commentend);
				URL[] urls = extractTrackLinks(comment);
				comment = PlainTextExtractor.extractPlainText(removeLinks(comment));

				if (outputverbose) System.out.print("COMMENT (on "+pageidx+") "+comment);

				if (urls.length > 0) {
					numcomments++;
					if (outputverbose) System.out.print(" -> URL "+urls[0].toString());
					// extract trackid
					int idstart = urls[0].toString().lastIndexOf("/")+1;
					String trackid = urls[0].toString().substring(idstart, urls[0].toString().length());
					// must be number
					try {
						Integer.parseInt(trackid);
					} catch (NumberFormatException nfe) {
						continue;
					}
					if (outputverbose) System.out.print(" -> TRACKID "+trackid);
					if (outputverbose) System.out.println("");

					File jsonfile = new File(jsondir.getAbsolutePath()+"/"+trackid+".json");
					if (!jsonfile.exists())
						continue;
					String jsoninfo;
					try {
						jsoninfo = TextTool.readInFile(jsonfile, "utf-8");
					} catch (Exception e1) {
						e1.printStackTrace();
						continue;
					}
					Entry entry = mapper.readValue(jsoninfo, new TypeReference<Entry>() {});
					String jsonid = entry.getId();

					if (entry != null && jsonid.equals(trackid)) {
						String jsonkey = "";
						if (entry.getKey() != null) {
							StandardKey stdkey = entry.getKey().getStandard();
							jsonkey = stdkey.getLetter().toLowerCase()+
									(stdkey.getFlat().booleanValue()?"b":"")+
									(stdkey.getSharp().booleanValue()?"#":"")+
									stdkey.getChord().replace("or", "");
						}

						HashSet<String> jsonbpms = new HashSet<String>();
						String jsonbpm = ""+entry.getBpm().intValue();
						jsonbpms.add(jsonbpm);

						String jsonmp3 = entry.getSampleUrl();
						String jsongenreslug = "";
						if (!entry.getGenres().isEmpty())
							jsongenreslug = entry.getGenres().get(0).getSlug();

						// DO THE KEY STUFF 
						if (!comment.contains("mixed in key") && !comment.contains("mixed-in-key")  && !comment.contains("mik")
								&& !comment.contains("melodyne") && !comment.contains("just ")) {

							Matcher notkey = wrong_key_pattern.matcher(comment);
							comment = notkey.replaceAll(" ");

							Matcher keym = key_pattern.matcher(comment);
							boolean foundkey = false;
							// remove beatport ground key truth and matches from title 
							String key_match = null;
							while (keym.find()) {
								if (foundkey) { // only unambiguous entries
									key_match = null;
									break;
								}
								key_match = keym.group().replace("or", "").replace("flat", "b").replace("sharp", "#").replace(" ", "");
								if (key_match.equals(jsonkey))
									continue;
								foundkey = true;
							}
							if (key_match != null) {
								System.out.println("KEY\t" + trackid + ";" + key_match + ";" + jsongenreslug  + ";" + jsonmp3 + ";" + jsoninfo);

								if (outputverbose) System.out.println("BEATPORT_KEY\t" + trackid + ";" + jsonkey);
								numkeys++;
							}
						}

						// DO THE BPM STUFF
						// get potential bpm matches from title
						// title pattern eg, "title":"Sinister (Staffan Linzatti Remix)",
						String title = entry.getTitle().toLowerCase().trim();
						Matcher json_m = bpm_pattern.matcher(title);
						while (json_m.find()) {
							String bpm_match = json_m.group().replace(",", ".").replace("=", "").trim();
							jsonbpms.add(bpm_match);
						}
						if (comment.contains("bpm")) { // run bpm matching

							for (Pattern notpat : wrong_bpm_patterns) {
								Matcher m = notpat.matcher(comment);
								comment = m.replaceAll(" ");
							}
							//if (outputverbose) System.out.print(" -> BPM candidates: ");

							Matcher m = bpm_pattern.matcher(comment);
							boolean foundbpm = false;
							// remove beatport ground bpm truth and matches from title 
							while (m.find()) {
								String bpm_match = m.group().replace(",", ".").replace("=", "").trim();
								float bpm_match_val;
								try {
									bpm_match_val = Float.parseFloat(bpm_match);
								} catch (NumberFormatException e) {
									continue;
								}

								if (jsonbpms.contains(bpm_match) || bpm_match_val < 40 || bpm_match_val > 250)
									continue;

								//if (outputverbose) System.out.println(bpm_match);
								//if (outputverbose) System.out.println("JSON " + jsoninfo);

								foundbpm = true;

								System.out.println("TEMPO\t" + trackid + ";" + bpm_match + ";" + jsongenreslug  + ";" + jsonmp3 + ";" + jsoninfo);

								if (outputverbose) System.out.println("BEATPORT_TEMPO\t" + trackid + ";" + jsonbpm);

								break;
							}
							if (foundbpm) numbpm++;
							//if (outputverbose) System.out.println("");
						}
					}
				}
				else {
					if (outputverbose) System.out.println("");
				}
			}
			pageidx++;
		}
		if (outputverbose) System.out.println("STAT\tno. comments total: "+totalnumcomments+" / with URL:"+numcomments+" / with BPM:"+numbpm+" / with KEY:"+numkeys);
	}

	private static URL[] extractTrackLinks(String content) {
		String lowercontent = content.toLowerCase();
		Vector<URL> urls = new Vector<URL>();

		int index = 0;
		while ((index = lowercontent.indexOf("<a ", index)) != -1) {
			if ((index = lowercontent.indexOf("href", index)) == -1) 
				break;
			if ((index = lowercontent.indexOf("=", index)) == -1) 
				break;
			index++;

			String remaining = content.substring(index);
			StringTokenizer st 
			= new StringTokenizer(remaining, "\t\n\r\">#");
			String strLink = st.nextToken();

			URL urlLink;
			try {
				urlLink = new URL(strLink);
			} catch (MalformedURLException e) {
				continue;
			}
			if (urlLink.toString().startsWith("http://www.beatport.com/track/"))
				urls.addElement(urlLink);
		}
		return urls.toArray(new URL[urls.size()]);
	}

	public static String removeLinks(String content) {
		String lowercontent = content.toLowerCase();
		StringBuffer noLinkContent = new StringBuffer();

		int contentStartIdx = 0;
		int index = 0;
		while ((index = lowercontent.indexOf("<a ", index)) != -1) {
			noLinkContent.append(content.substring(contentStartIdx, index)+" ");
			index = lowercontent.indexOf("</a>", index);
			if (index != -1) {
				contentStartIdx = index+4;
			}
			else {
				contentStartIdx = content.length();
				break;
			}
		}
		noLinkContent.append(content.substring(contentStartIdx, content.length()));
		return noLinkContent.toString();
	}
	
	public static String retrieve(URL url) {
		String useragent = "Mozilla/5.0 (Windows; U; Windows NT 5.1; en-US; rv:1.8.1.3) Gecko/20070309 Firefox/2.0.0.3";
		String encoding = "utf-8";

		StringBuilder sb = new StringBuilder();

		try {
			char[] buffer = new char[2048];

			URLConnection uc = url.openConnection();
			uc.setConnectTimeout(1000);
			uc.setReadTimeout(2000);
			uc.addRequestProperty("User-Agent", useragent);
			InputStream is = uc.getInputStream();
			InputStreamReader isr = new InputStreamReader(is, Charset.forName(encoding));

			int len;
			while ((len = isr.read(buffer)) != -1) {
				sb.append(buffer, 0, len);
			}
		} catch (Exception e) {
			return null;
		}

		return sb.toString();
	}



	public static void main(String[] args) {
		try {
			//downloadForumPages();
			extractGroundTruth();
		} catch (Exception e) {
			e.printStackTrace();
		}
	}


}
